package com.brilliance.eibs.coder.impl;

import com.brilliance.eibs.coder.AsymmetryCoder;

import javax.crypto.KeyAgreement;
import javax.crypto.SecretKey;
import javax.crypto.interfaces.DHPrivateKey;
import javax.crypto.interfaces.DHPublicKey;
import javax.crypto.spec.DHParameterSpec;
import java.security.*;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.HashMap;
import java.util.Map;

/**
 * @author hujun
 *
 */
public class DHCoder
    extends AsymmetryCoder
{

  private static final String KEY_ALGORITHM = "DH";
  //由于生成长度超128bits的AES密钥需要
  //单独从Oracle官网下载对应JDK版本的Java Cryptography Extension (JCE) Unlimited Strength Jurisdiction Policy Files文件
  //应此如果没有以上版本JDK，不能使用AES加密
  public static final String  AES           = "AES";
  public static final String  DES           = "DES";
  public static final String  DESEDE        = "DESede";

  //key_size必须为512-1024 64被倍数 默认1024
  private static final int    KEY_SIZE      = 512;

  public static Map<String, Object> initKey(int keySize)
      throws NoSuchAlgorithmException
  {
    return generateKey(null, keySize);
  }

  public static Map<String, Object> initKey()
      throws NoSuchAlgorithmException
  {
    return initKey(KEY_SIZE);
  }

  public static Map<String, Object> generateKey(String strKey)
      throws NoSuchAlgorithmException
  {
    return generateKey(strKey, KEY_SIZE);
  }

  public static Map<String, Object> generateKey(String strKey, int keySize)
      throws NoSuchAlgorithmException
  {
    return generateKey(strKey, keySize, KEY_ALGORITHM);
  }

  public static Map<String, Object> initKeyByPublicKey(byte[] key)
      throws NoSuchAlgorithmException, InvalidKeySpecException, InvalidAlgorithmParameterException
  {
    return generateKeyByPublicKey(key, null);
  }

  public static Map<String, Object> generateKeyByPublicKey(byte[] key, String strKey)
      throws NoSuchAlgorithmException, InvalidKeySpecException, InvalidAlgorithmParameterException
  {
    X509EncodedKeySpec x509KeySpec = new X509EncodedKeySpec(key);
    KeyFactory keyFactory = KeyFactory.getInstance(KEY_ALGORITHM);
    PublicKey pubKey = keyFactory.generatePublic(x509KeySpec);
    DHParameterSpec dhParamSpec = ((DHPublicKey) pubKey).getParams();
    KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance(pubKey.getAlgorithm());
    keyPairGenerator.initialize(dhParamSpec, getSecureRandom(strKey));
    KeyPair keyPair = keyPairGenerator.genKeyPair();
    DHPublicKey publicKey = (DHPublicKey) keyPair.getPublic();
    DHPrivateKey privateKey = (DHPrivateKey) keyPair.getPrivate();
    Map<String, Object> keyMap = new HashMap<String, Object>(2);
    keyMap.put(PUBLIC_KEY, publicKey);
    keyMap.put(PRIVATE_KEY, privateKey);
    return keyMap;
  }

  /*
   * 具体的利用本地密钥加密解密可以使用其他Coder中的加密解密方法，而不在本类中描述
   */
  /*  public static byte[] encrypt(byte[] data, byte[] key)
        throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException
    {
      return encrypt(data, key, AES);
    }

    public static byte[] encrypt(byte[] data, byte[] key, String algorithm)
        throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException
    {
      SecretKey secretKey = new SecretKeySpec(key, algorithm);
      Cipher cipher = Cipher.getInstance(secretKey.getAlgorithm());
      cipher.init(Cipher.ENCRYPT_MODE, secretKey);
      return cipher.doFinal(data);
    }

    public static byte[] decrypt(byte[] data, byte[] key)
        throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException
    {
      return decrypt(data, key, AES);
    }
  
  public static byte[] decrypt(byte[] data, byte[] key, String algorithm)
      throws NoSuchAlgorithmException, NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException
  {
    SecretKey secretKey = new SecretKeySpec(key, algorithm);
    Cipher cipher = Cipher.getInstance(secretKey.getAlgorithm());
    cipher.init(Cipher.DECRYPT_MODE, secretKey);
    return cipher.doFinal(data);
  }
  */
  public static byte[] getSecretKey(byte[] publicKey, byte[] privateKey)
      throws NoSuchAlgorithmException, InvalidKeySpecException, InvalidKeyException
  {
    return getSecretKey(publicKey, privateKey, AES);
  }

  public static byte[] getSecretKey(byte[] publicKey, byte[] privateKey, String algorithm)
      throws NoSuchAlgorithmException, InvalidKeySpecException, InvalidKeyException
  {
    KeyFactory keyFactory = KeyFactory.getInstance(KEY_ALGORITHM);
    X509EncodedKeySpec x509KeySpec = new X509EncodedKeySpec(publicKey);
    PublicKey pubKey = keyFactory.generatePublic(x509KeySpec);
    PKCS8EncodedKeySpec pkcs8KeySpec = new PKCS8EncodedKeySpec(privateKey);
    PrivateKey priKey = keyFactory.generatePrivate(pkcs8KeySpec);
    KeyAgreement keyAgree = KeyAgreement.getInstance(keyFactory.getAlgorithm());
    keyAgree.init(priKey);
    keyAgree.doPhase(pubKey, true);
    SecretKey secretKey = keyAgree.generateSecret(algorithm);
    return secretKey.getEncoded();
  }

}
