package com.brilliance.eibs.core.service.instance.plugin;

import com.brilliance.eibs.core.service.Context;
import com.brilliance.eibs.util.LogUtil;
import org.apache.commons.collections.map.CaseInsensitiveMap;
import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.sql.*;
import java.util.Date;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * jdbc操作执行器
 *
 * @author hujun
 */
public final class JdbcExecutor {

    private Connection conn;

    private PreparedStatement pstmt;
    private CallableStatement cstmt;
    private Context context;
    private ResultSet rs;
    Logger log = LogUtil.getLogger(context);
    public static String RESULTSET_KEY = "procedure_result_for_inner_use";

    public JdbcExecutor(Context ctx, Connection conn) throws SQLException {
        this.context = ctx;
        this.conn = conn;

        log.debug("Create a executor for jdbc connection.");
        if (null == this.conn) {
            log.error("Get a jdbc connection failed.");
            throw new RuntimeException("Get a jdbc connection failed.");
        }
    }

    public String getProductName() throws SQLException {
        return conn.getMetaData().getDatabaseProductName();
    }

    public void commit() throws SQLException {
        log.debug("Database commit.");
        conn.commit();
    }

    public void setSavePoint(String name) throws SQLException {
        conn.setSavepoint(name);
    }

    public void rollBack() throws SQLException {
        log.debug("Database rollback.");
        conn.rollback();
    }

    public void close() throws SQLException {
        log.debug("Close Connection.");
        if (null != rs) rs.close();
        rs = null;
        if (null != pstmt) pstmt.close();
        pstmt = null;
        if (null != conn) {
            if (!conn.getAutoCommit()) {
                conn.setAutoCommit(true);
                conn.close();
            }
        }
        conn = null;
    }

    public void setAutoCommit(boolean autoCommit) throws SQLException {
        conn.setAutoCommit(autoCommit);
    }

    public boolean isClosed() throws SQLException {
        if (conn == null) return true;
        return conn.isClosed();
    }

    public void closeWithoutConn() throws SQLException {
        log.debug("Close resultSet and preparedStatement.");
        if (null != rs) rs.close();
        rs = null;
        if (null != pstmt) pstmt.close();
        pstmt = null;
    }

    private ResultSet executeQueryRS(PreparedStatement pstmt) throws SQLException {
        return pstmt.executeQuery();
    }

    private List<Map<String, Object>> executeQuery(PreparedStatement pstmt) throws SQLException {
        List<Map<String, Object>> list = null;
        list = new ArrayList<Map<String, Object>>();
        rs = pstmt.executeQuery();
        ResultSetMetaData rsmd = pstmt.getMetaData();
        int mdCount = rsmd.getColumnCount();
        log.debug("Columne count is " + mdCount);
        while (rs.next()) {
            log.debug("Return result sets, the content of a new record:");
            Map<String, Object> map = new CaseInsensitiveMap();
            for (int i = 1; i <= mdCount; i++) {
                Object col = rs.getObject(i);
                if (col instanceof Blob) {
                    Blob b = (Blob) col;
                    byte[] buf = new byte[(int) b.length()];
                    try {
                        IOUtils.readFully(b.getBinaryStream(), buf);
                    } catch (IOException e) {
                    }
                    col = buf;
                    map.put(rsmd.getColumnLabel(i), col);
                } else if (col instanceof Clob) {
                    Clob clob = (Clob) col;
                    Reader is = clob.getCharacterStream();
                    BufferedReader br = new BufferedReader(is);
                    String s;
                    try {
                        s = br.readLine();
                        StringBuffer sb = new StringBuffer();
                        while (s != null) {
                            sb.append(s);
                            s = br.readLine();
                        }
                        map.put(rsmd.getColumnLabel(i), sb.toString());
                    } catch (IOException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }

                } else {
                    map.put(rsmd.getColumnLabel(i), col);
                }


            }
            list.add(map);
        }
        return list;
    }

    public Map<String, String> executeQueryForCodeTables(String sql) throws SQLException {
        Map<String, String> map = new HashMap<String, String>();
        try {
            pstmt = conn.prepareStatement(sql);
            rs = pstmt.executeQuery();
            while (rs.next()) {
                map.put(rs.getObject(1).toString(), rs.getObject(2).toString());
            }
        } finally {
            closeWithoutConn();
        }
        return map;

    }

    public ResultSet executeQueryRS(String sql) throws SQLException {
        log.debug("Execute a query, the sql is " + sql);
        pstmt = conn.prepareStatement(sql, ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.CONCUR_READ_ONLY);
        return executeQueryRS(pstmt);
    }

    public List<Map<String, Object>> executeQuery(String sql) throws SQLException {
        log.debug("Execute a query, the sql is " + sql);
        pstmt = conn.prepareStatement(sql);
        try {
            return executeQuery(pstmt);
        } finally {
            closeWithoutConn();
        }
    }

    public Map<String, Object> executeQueryForMap(String sql) throws SQLException {
        List<Map<String, Object>> list = executeQuery(sql);
        if (null != list && !list.isEmpty())
            return list.get(0);
        else
            return null;
    }

    public ResultSet executeQueryRS(String sql, Object[] params) throws SQLException {
        // log.debug("Execute a query with parameters, the sql is [" + sql +
        // "]");
        // 新增sql拼接处理
        params = (Object[]) dealSql(sql, params).get(1);
        pstmt = conn.prepareStatement(sql, ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.CONCUR_READ_ONLY);
        // pstmt = conn.prepareStatement(sql);
        // pstmtSetParams(params);
        pstmtSetParams(sql, params);
        return executeQueryRS(pstmt);
    }

    public List<?> dealSql(String sql, Object[] params) {
        List list = new ArrayList();
        if (params != null && params.length == 1 && params[0] instanceof Map) {
            Map<String, Object> map = (Map<String, Object>) params[0];
            Object[] o = null;
            if (map.size() > 0) {
                o = new Object[map.size()];
                int i = 0;
                for (String key : map.keySet()) {
                    /*
                     * if (!key.contains(":")) { sql = sql + " AND " + key + "=?"; } else { String
                     * key1=key.split(":")[0]; String sysbol=key.split(":")[1]; sql=sql+" AND "
                     * +key1+sysbol+"?"; }
                     */
                    o[i] = map.get(key);
                    i++;
                }
                params = o;
            }
        }
        list.add(sql);
        list.add(params);
        return list;
    }

    public List<Map<String, Object>> executeQuery(String sql, Object[] params) throws SQLException {
        List<Map<String, Object>> list = new ArrayList<Map<String, Object>>();
        // 新增sql拼接处理
        params = (Object[]) dealSql(sql, params).get(1);
        log.debug("Execute a query with parameters, the sql is [" + sql + "]");
        pstmt = conn.prepareStatement(sql);
        // pstmtSetParams(params);
        pstmtSetParams(sql, params);
        try {
            list = executeQuery(pstmt);
            return list;
        } finally {
            closeWithoutConn();
        }
    }

    // private void pstmtSetParams(Object[] params)
    // throws SQLException
    // {
    // if (null != params && params.length != 0)
    // {
    // for (int i = 0; i < params.length; i++)
    // {
    // if (params[i] instanceof java.util.Date)
    // {
    // params[i] = new java.sql.Date(((Date) params[i]).getTime());
    // }
    // pstmt.setObject(i + 1, params[i]);
    // }
    // }
    // }

    private void pstmtSetParams(String sql, Object[] params) throws SQLException {
        StringBuffer returnSQL = new StringBuffer();
        String[] subSQL = sql.split("\\?");
        // 修改by gechengyang
        if (null != params && params.length != 0 && !(params[0] instanceof Map))
        // if (null != params && params.length != 0)
        {
            for (int i = 0; i < params.length; i++) {
                if (params[i] instanceof Date) {
                    // params[i] = new java.sql.Date(((Date)
                    // params[i]).getTime());
                    /**
                     * java.util.Date是不能直接插入数据库中的，虽然它包含了日期和时分秒。java.sql. Date可以直接插入数据库
                     * ，但是它只有日期而没有时间；java.sql.Time可以直接插入数据库，但是它只有时间， 没有日期；以上这些往往不能满足我们的需要。
                     * 如果我们要插入日期和时间，那么就要相办法来使用java.sql.Timestamp来向数据库插入
                     **/
                    params[i] = new Timestamp(((Date) params[i]).getTime());
                }
                pstmt.setObject(i + 1, params[i]);
                if (params[i] instanceof String)
                    returnSQL.append(subSQL[i]).append(" '").append(params[i]).append("' ");
                else
                    returnSQL.append(subSQL[i]).append(params[i]);
            }
            if (params.length == (subSQL.length - 1)) {
                returnSQL.append(subSQL[subSQL.length - 1]);
            }
            log.info("The excuete sql is[" + returnSQL.toString() + "].");
        }
    }

    private void cstmtSetParams(String sql, Object[][] params) throws SQLException {
        StringBuffer returnSQL = new StringBuffer();
        String[] subSQL = sql.split("\\?");
        if (null != params && params.length != 0) {
            for (int i = 0; i < params.length; i++) {
                if (params[i][2] instanceof Date) {
                    params[i][2] = new java.sql.Date(((Date) params[i][3]).getTime());
                }
                if (params[i][1].equals("IN") || params[i][1].equals("INOUT"))
                    cstmt.setObject(i + 1, params[i][2]);
                if (params[i][1].equals("OUT") || params[i][1].equals("INOUT"))
                    cstmt.registerOutParameter(i + 1, (Integer) params[i][0]);
                returnSQL.append(subSQL[i]);
                if (params[i][2] == null)
                    returnSQL.append(" null ");
                else if (params[i][2] instanceof String)
                    returnSQL.append(" '").append(params[i][2]).append("' ");
                else
                    returnSQL.append(params[i][2]);
            }
            if (params.length == (subSQL.length - 1)) {
                returnSQL.append(subSQL[subSQL.length - 1]);
            }
            log.info("The excuete sql is[" + returnSQL.toString() + "].");
        } else {
            log.info("The excuete sql is[" + sql + "].");
        }
    }

    @Deprecated
    public List<Map<String, Object>> executeNamedQuery(String sql, Map<String, Object> params)
            throws SQLException {
        Object[] array = map2Array(params, sql);
        return executeQuery(replaceForSqlParm(sql), array);
    }

    @SuppressWarnings("rawtypes")
    private Object[] map2Array(Map<String, Object> params, String sql) {
        Object[] array = null;
        if (null != params && params.size() != 0) {
            List<String> paramNames = getGroupsForSqlParm(sql);
            array = new Object[paramNames.size()];
            for (int i = 0; i < paramNames.size(); i++) {
                Object value = params.get(paramNames.get(i));
                if (null == value) {
                    log.error("Mismatch parameter exists.");
                    throw new RuntimeException("Mismatch parameter exists.");
                } else {
                    if (value instanceof Date) {
                        value = new java.sql.Date(((Date) value).getTime());
                    }
                    if (value instanceof List) {
                        value = ((List) value).get(0);
                    }
                    array[i] = value;
                }
            }
        }
        return array;
    }

    public int executeUpdate(String sql, Object[] params) throws SQLException {
        // log.debug("Execute a update with parameters, the sql is [" + sql +
        // "]");
        pstmt = conn.prepareStatement(sql);
        // pstmtSetParams(params);
        pstmtSetParams(sql, params);
        return executeUpdate(pstmt);
    }


    public int executeUpdate(String sql) throws SQLException {
        // log.debug("Execute a update, the sql is [" + sql + "]");
        pstmt = conn.prepareStatement(sql);
        return executeUpdate(pstmt);
    }


    private int executeUpdate(PreparedStatement pstmt) throws SQLException {
        try {
            return pstmt.executeUpdate();
        } finally {
            closeWithoutConn();
        }
    }

    public int executeNamedUpdate(String sql, Map<String, Object> params) throws SQLException {
        Object[] array = map2Array(params, sql);
        return executeUpdate(replaceForSqlParm(sql), array);
    }

    public boolean executeProcedure(CallableStatement cstmt) throws SQLException {
        return cstmt.execute();
    }

    public ResultSet executeProcedureRS(CallableStatement cstmt) throws SQLException {
        return cstmt.executeQuery();
    }

    public boolean executeProcedure(String sql) throws SQLException {
        cstmt = conn.prepareCall(sql);
        return executeProcedure(cstmt);
    }

    public Map<String, Object> executeProcedure(String sql, Object[][] params) throws SQLException {
        Map<String, Object> map = new HashMap<String, Object>();
        cstmt = conn.prepareCall(sql, ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.CONCUR_READ_ONLY);
        cstmtSetParams(sql, params);
        // boolean b = executeProcedure(cstmt);
        // System.out.println("result:" + b);
        ResultSet resultSet = cstmt.executeQuery();
        for (int i = 0; i < params.length; i++) {
            if (params[i][1].equals("OUT") || params[i][1].equals("INOUT")) {
                Object object = cstmt.getObject(i + 1);
                if (object instanceof Blob) {
                    Blob b = (Blob) object;
                    byte[] buf = new byte[(int) b.length()];
                    try {
                        IOUtils.readFully(b.getBinaryStream(), buf);
                    } catch (IOException e) {
                    }
                    object = buf;
                }

                map.put((String) params[i][3], object);
            }
        }
        // ResultSet resultSet = cstmt.getResultSet();
        if (resultSet != null) {
            map.put(RESULTSET_KEY, resultSet);
        }
        return map;
    }

    public ResultSet executeProcedureRS(String sql, Object[][] params) throws SQLException {
        cstmt = conn.prepareCall(sql, ResultSet.TYPE_SCROLL_INSENSITIVE,
                ResultSet.CONCUR_READ_ONLY);
        cstmtSetParams(sql, params);
        return cstmt.executeQuery();
    }

    /**
     * 提取sql当中的命名参数
     *
     * @param target 目标表达式，目前仅为sql
     * @return 返回sql当中的参数列表
     * @author hujun
     */
    private List<String> getGroupsForSqlParm(String target) {
        List<String> groups = new ArrayList<String>();
        Pattern p = getDefaultPattern();
        Matcher matcher = p.matcher(target);// p.split()就是string.split的实现
        while (matcher.find()) {
            String paramName = matcher.group().substring(1);
            if (!groups.contains(paramName)) groups.add(paramName);
        }
        return groups;
    }

    /**
     * 将sql当中的命名参数替换为“？”
     *
     * @param target 目标表达式，目前仅为sql
     * @return 返回替换后的sql
     * @author hujun
     */
    private String replaceForSqlParm(String target) {
        Pattern p = getDefaultPattern();
        Matcher matcher = p.matcher(target);// p.split()就是string.split的实现
        matcher.find();
        return matcher.replaceAll("?");
    }

    /**
     * @return 获取命名参数正则对应的pattern
     * @author hujun
     */
    private Pattern getDefaultPattern() {
        String exp = ":[^0-9_](_|[a-zA-Z0-9\\[\\]\\'.\\(\\)])*";// 头字母不能为数字和下划线,支持字母数字和下划线
        // String exp = ":[^0-9](_|[a-zA-Z0-9])*";//头字母不能为数字,支持字母数字和下划线
        Pattern p = Pattern.compile(exp);
        return p;
    }

    public PreparedStatement getPrepareStatement(String sql) throws SQLException {
        return conn.prepareStatement(sql);
    }
}
