package com.brilliance.eibs.core.service.instance.plugin;

import com.brilliance.eibs.core.frame.ISocketProxy;
import com.brilliance.eibs.extension.core.IClient;
import org.apache.mina.core.buffer.IoBuffer;
import org.apache.mina.core.session.IoSession;

import java.io.IOException;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class MinaSocketProxy implements ISocketProxy {

	private IoSession session;
	private volatile Object message;
	private String encoding = "utf-8";
	private int receiveTimeout=2000;

	// 由于mina框架是异步处理，但是接口平台相关socket配置是基于同步处理，所以加入锁机制，确保在配置获取返回信息的时候，mina框架已经接收到返回信息
	private Lock lock = new ReentrantLock();
	private Condition condition = lock.newCondition();
	private IClient client;

	public MinaSocketProxy(IoSession session, Object message,String charset,int timeout, IClient client) {
		this.session = session;
		this.message = message;
		this.encoding=charset;
		this.receiveTimeout=timeout;
		this.client=client;
	}

	public MinaSocketProxy(String charset, int timeout, IClient client) {
		this.encoding=charset;
		this.receiveTimeout=timeout;
		this.client=client;
	}
	
	public MinaSocketProxy() {
	}


	public void setSession(IoSession session) {
		this.session = session;
	}

	public void setMessage(Object message) {
		lock.lock();
        try {
        	this.message = message;
            if (condition != null) {
            	condition.signal();
            }
        } finally {
            lock.unlock();
        }
	}

	public void setReceiveTimeout(int receiveTimeout) {
		this.receiveTimeout = receiveTimeout;
	}

	private void suspend() throws IOException {
		if (!isDone()) {
            long start = System.currentTimeMillis();
            lock.lock();
            try {
                while (!isDone()) {
                	condition.await(this.receiveTimeout, TimeUnit.MILLISECONDS);
                    if (isDone() || System.currentTimeMillis() - start > receiveTimeout) {
                        break;
                    }
                }
            } catch (InterruptedException e) {
                throw new RuntimeException(e);
            } finally {
                lock.unlock();
            }
            if (!isDone()) {
            	close();
            	throw new SocketTimeoutException("读取数据超时，单位毫秒"+this.receiveTimeout);
            }
        }
	}

	private boolean isDone() {
		return this.message != null;
	}

	@Override
	public void close() {
		this.client.close();
	}

	@Override
	public void setEncoding(String encoding) {
		this.encoding = encoding;
	}

	@Override
	public String getDataStream(int len) throws IOException {
		String s = new String(getDataStreamBytes(len), encoding);
		return s;
	}

	@Override
	public Object[] readDataByBuffer() throws IOException {
		throw new UnsupportedOperationException("MinaSocketProxy does not support this method.");
	}

	@Override
	public byte[] getDataStreamBytes(int len) throws IOException {
		suspend();
		if(message instanceof IoBuffer){
			IoBuffer in = (IoBuffer) message;
			byte[] buf = new byte[len];
			in.get(buf);
			return buf;
		}else{
			throw new UnsupportedOperationException("返回得信息不支持字节流方式读取.对象是=="+message.getClass());
		}
		
	}

	@Override
	public String getStrBySocketStream() throws IOException {
		throw new UnsupportedOperationException("MinaSocketProxy does not support this method.");
	}

	@Override
	public void send(Object sendMsg) throws IOException {
		if (sendMsg == null) {
			return;
		}
		session.write(sendMsg);
	}

	@Override
	public void skip(long len) throws IOException {
		getDataStreamBytes((int) len);
	}

	@Override
	public boolean isConnected() {
		return !session.isClosing() && session.isConnected();
	}

	@Override
	public Socket getSocket() {
		// TODO Auto-generated method stub
		throw new UnsupportedOperationException("MinaSocketProxy does not support this method.");
	}

	@Override
	public Object getObject() throws IOException {
		suspend();
		if(message instanceof IoBuffer){
			IoBuffer in = (IoBuffer) message;
			byte[] buf = new byte[in.remaining()];
			in.get(buf);
			return buf;
		}else{
			return this.message;
		}
	}

}
