package com.brilliance.eibs.factory.parser.impl;

import com.brilliance.eibs.core.exception.InterfaceException;
import com.brilliance.eibs.core.model.IFuncDef;
import com.brilliance.eibs.core.model.IFunctionDef;
import com.brilliance.eibs.core.model.impl.FuncDef;
import com.brilliance.eibs.core.model.impl.FunctionDef;
import com.brilliance.eibs.core.model.impl.FunctionDef.Input;
import com.brilliance.eibs.core.model.impl.ReflectDef;
import com.brilliance.eibs.core.service.instance.impl.AbsFilter;
import com.brilliance.eibs.factory.BeanFactory.FunctionKey;
import com.brilliance.eibs.factory.FactoryManager;
import com.brilliance.eibs.util.StringUtil;
import com.brilliance.eibs.util.XmlUtil;
import org.dom4j.Element;

import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author hujun
 *
 *         自定义函数配置解析
 */
public class FunctionParser {
    private XmlParser parser;

    public FunctionParser(XmlParser parser) {
        this.parser = parser;
    }

    /**
     * 初始化自定义函数列表
     */
    public Map<FunctionKey, IFunctionDef> getFunctionMap(List<Element> rootNodes) {
        Map<FunctionKey, IFunctionDef> functionMap = new HashMap<FunctionKey, IFunctionDef>();
        for (Element rootNode : rootNodes) {
            String ref = parser.getAttribute(rootNode, "ref");
            if (!StringUtil.isEmpty(ref)) {
                URL url = XmlUtil.getUrl(ref);
                if (url != null) {
                    Element rootElement = XmlUtil.getDocument(url).getRootElement();
                    functionMap.putAll(getFunctionMap(rootElement));
                } else
                    throw new InterfaceException("00006", "can not find resource path: " + ref);
            } else
                functionMap.putAll(getFunctionMap(rootNode));
        }
        return functionMap;
    }

    /**
     * 获取节点下的所有自定义函数定义，放入Map
     * 
     * @param rootNode
     * @return
     */
    public Map<FunctionKey, IFunctionDef> getFunctionMap(Element rootNode) {
        Map<FunctionKey, IFunctionDef> map = new HashMap<FunctionKey, IFunctionDef>();
        List<Element> functionNodes = rootNode.elements("function");
        for (Element functionNode : functionNodes) {
            String funcName = parser.getRequiredAttribute(functionNode, "name");
            List<Element> inputNodes = functionNode.elements("input");
            map.put(buildBaseFunctionKey(funcName, inputNodes),
                    getFunctionDef(functionNode, inputNodes));
        }
        map.putAll(getReflectFunction(rootNode));
        return map;
    }

    /**
     * @param name
     * @param inputNodes
     * @return
     *
     *         构建标准自定义函数索引
     */
    private FunctionKey buildBaseFunctionKey(String name, List<Element> inputNodes) {
        Class<?>[] parameterTypes = null;
        if (inputNodes != null) {
            parameterTypes = new Class[inputNodes.size()];
            String inputType = null;
            for (int i = 0; i < inputNodes.size(); i++) {
                inputType = parser.getAttribute(inputNodes.get(i), "type");
                parameterTypes[i] =
                        StringUtil.isEmpty(inputType) ? Object.class : obtainClassType(inputType);
            }
        }
        return buildFunctionKey(name, parameterTypes);
    }

    /**
     * @param inputType
     * @return
     *
     *         返回非空type的非数组class类型
     */
    private Class<?> NonemptyClass(String inputType) {
        inputType = inputType.trim();
        Class<?> returnCls = AbsFilter.typeMap.get(inputType);
        try {
            return returnCls == null ? Class.forName(inputType) : returnCls;
        } catch (ClassNotFoundException e) {
            throw new InterfaceException("00422", e);
        }
    }

    /**
     * @param type
     * @return
     *
     *         返回非空type的class类型，支持数组类型的返回
     */
    private Class<?> obtainClassType(String type) {
        type = type.trim();
        if (isArray(type)) {
            String arrayClasstype = type.substring(0, type.indexOf("[]"));
            Class<?> arrayClass = NonemptyClass(arrayClasstype);
            Pattern p = Pattern.compile("\\[\\]");
            Matcher m = p.matcher(type);
            int index = 0;
            while (m.find())
                index++;
            return java.lang.reflect.Array.newInstance(arrayClass, new int[index]).getClass();
        }
        return NonemptyClass(type);
    }

    private boolean isArray(String type) {
        return type.trim().matches("[\\w\\.]+(\\[\\])+$");
    }

    /**
     * @param name
     * @param type
     * @return
     *
     *         构建反射方法的索引
     */
    private FunctionKey buildReflectFunctionKey(String name, String type) {
        if (StringUtil.isEmpty(type)) return buildFunctionKey(name, new Class[0]);
        String[] clsstr = StringUtil.split(",", type);
        Class<?>[] parameterTypes = new Class[clsstr.length];
        for (int i = 0; i < clsstr.length; i++)
            parameterTypes[i] =
                    StringUtil.isEmpty(clsstr[i]) ? Object.class : obtainClassType(clsstr[i]);
        return buildFunctionKey(name, parameterTypes);
    }

    /**
     * @param name
     * @param parameterTypes
     * @return
     *
     *         构建自定义函数索引
     */
    public static FunctionKey buildFunctionKey(String name, Class<?>[] parameterTypes) {
        return FactoryManager.getBeanFactory().new FunctionKey(name, parameterTypes);
    }

    /**
     * @param doc
     *
     *        初始化反射自定义函数列表
     */
    private Map<FunctionKey, IFunctionDef> getReflectFunction(Element rootNode) {
        Map<FunctionKey, IFunctionDef> map = new HashMap<FunctionKey, IFunctionDef>();
        for (Element functionNode : (List<Element>) rootNode.elements("reflect")) {
            for (Element reflectNode : (List<Element>) functionNode.elements()) {
                if (ReflectDef.CONSTRUCTOR_TYPE.equalsIgnoreCase(reflectNode.getName())
                        || ReflectDef.FIELD_TYPE.equalsIgnoreCase(reflectNode.getName())
                        || ReflectDef.METHOD_TYPE.equalsIgnoreCase(reflectNode.getName())) {
                    ReflectDef reflectDef = getReflectDef(reflectNode);
                    map.put(buildReflectFunctionKey(
                            StringUtil.isEmpty(reflectDef.getAlias())
                                    ? reflectDef.getName()
                                    : reflectDef.getAlias(),
                            reflectDef.getType()), reflectDef);
                }
            }
        }
        return map;
    }


    public Map<FunctionKey, IFuncDef> getFuncMap(List<Element> rootNodes) {
        Map<FunctionKey, IFuncDef> map = new HashMap<FunctionKey, IFuncDef>();
        for (Element rootNode : rootNodes) {
            String ref = parser.getAttribute(rootNode, "ref");
            if (!StringUtil.isEmpty(ref)) {
                URL url = XmlUtil.getUrl(ref);
                if (url != null) {
                    Element rootElement = XmlUtil.getDocument(url).getRootElement();
                    map.putAll(getFunc(rootElement));
                } else
                    throw new InterfaceException("00006", "can not find resource path: " + ref);
            } else
                map.putAll(getFunc(rootNode));
        }
        return map;
    }


    /**
     * 获取func节点相关的属性值
     **/
    public Map<FunctionKey, IFuncDef> getFunc(Element rootNode) {
        Map<FunctionKey, IFuncDef> map = new HashMap<FunctionKey, IFuncDef>();
        List<Element> functionNodes = rootNode.elements("func");
        for (Element functionNode : functionNodes) {
            String namespace = parser.getAttribute(functionNode, "namespace");
            String cls = parser.getAttribute(functionNode, "class");
            String stc = parser.getAttribute(functionNode, "static");

            IFuncDef functionDef = new FuncDef();
            functionDef.setNamespace(namespace);
            functionDef.setCls(cls);
            functionDef.setStc(stc);
            map.put(FactoryManager.getBeanFactory().new FunctionKey(namespace + cls + stc, null),
                    functionDef);
        }
        return map;
    }

    /**
     * @param functionNode
     * @return
     * 
     *         获得反射方法的static属性，以子节点上的static方法为准
     */
    private boolean getReflectStaticSts(Element functionNode) {
        String isStatic = parser.getAttribute(functionNode, "static");
        return Boolean.valueOf(StringUtil.isEmpty(isStatic)
                ? parser.getAttribute(functionNode.getParent(), "static")
                : isStatic);
    }

    /**
     * @param functionNode
     * @return
     * 
     *         获得反射自定义函数模型实例
     */
    private ReflectDef getReflectDef(Element functionNode) {
        return new ReflectDef(functionNode.getName().trim(),
                parser.getAttribute(functionNode, "type"), getReflectName(functionNode),
                parser.getRequiredAttribute(functionNode.getParent(), "class"),
                parser.getAttribute(functionNode, "error"),
                parser.getAttribute(functionNode, "alias"), getReflectStaticSts(functionNode))
                        .resetType();
    }

    /**
     * 从reflect字节中获取需要的name参数，考虑如果构造函数无需配置name直接从包名中获得
     * 
     * @param functionNode
     * @return
     */
    private String getReflectName(Element functionNode) {
        String name = parser.getAttribute(functionNode, "name");
        return StringUtil.isEmpty(name) ? getClassNameFromPackage(functionNode) : name;
    }

    /**
     * 从包名中获取类名
     * 
     * @param functionNode
     * @return
     */
    private String getClassNameFromPackage(Element functionNode) {
        if (!ReflectDef.CONSTRUCTOR_TYPE.equalsIgnoreCase(functionNode.getName()))
            throw new InterfaceException("00005", "attribute [name] of node ["
                    + functionNode.getPath() + "] should not be empty.");
        String clazz = parser.getRequiredAttribute(functionNode.getParent(), "class");
        return clazz.substring(clazz.lastIndexOf(".") + 1, clazz.length());
    }

    /**
     * @param functionNode
     * @param inputNodes
     * @return
     * 
     *         获得标准自定义函数模型实例
     */
    private IFunctionDef getFunctionDef(Element functionNode, List<Element> inputNodes) {
        FunctionDef functionDef = new FunctionDef();
        functionDef.setName(parser.getRequiredAttribute(functionNode, "name"));
        functionDef.setTag(parser.getAttribute(functionNode, "tag"));
        functionDef.setError(parser.getAttribute(functionNode, "error"));
        for (Element inputNode : inputNodes) {
            Input input = functionDef.new Input();
            input.setTag(parser.getRequiredAttribute(inputNode, "tag"));
            input.setType(parser.getAttribute(inputNode, "type"));
            functionDef.addInput(input);
            functionNode.remove(inputNode);
        }
        functionDef.setExecuteList(parser.getFilterSubList(functionDef, functionNode));
        return functionDef;
    }
}
