package com.brilliance.eibs.server.link_v2_0;

import org.apache.commons.io.IOUtils;
import org.slf4j.Logger;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class SimplexLink
    implements Link
{
  private int              timeout  = 60000;
  private volatile boolean isReady  = false;
  private long             interval = 5000;

  private LinkData data;
  private Socket   send;
  private Socket   receive;

  private boolean isReceiveFirst;
  private boolean isReconnect;

  private Heartbeat     heartbeat;
  private ReceiveDetect receiveDetect;

  private ExecutorService executorService;

  private Lock lock = new ReentrantLock();

  private Logger logger;

  @SuppressWarnings("unused")
  private SimplexLink()
  {}

  public SimplexLink( LinkData data, Logger logger )
  {
    this(data, false, true, logger);
  }

  public SimplexLink( LinkData data, boolean isReceiveFirst, boolean isReconnect, Logger logger )
  {
    this.data = data;
    this.isReceiveFirst = isReceiveFirst;
    this.isReconnect = isReconnect;
    this.logger = logger;
  }

  public Logger getLogger()
  {
    return logger;
  }

  public boolean isReceiveFirst()
  {
    return isReceiveFirst;
  }

  public Heartbeat getHeartbeat()
  {
    return heartbeat;
  }

  public void setHeartbeat(Heartbeat heartbeat)
  {
    this.heartbeat = heartbeat;
  }

  public Lock getLock()
  {
    return lock;
  }

  public void setReceiveDetect(ReceiveDetect receiveDetect)
  {
    this.receiveDetect = receiveDetect;
  }

  public ReceiveDetect getReceiveDetect()
  {
    return receiveDetect;
  }

  private void startDetect()
  {
    if (null == executorService)
      executorService = Executors.newFixedThreadPool(2);
    if (null != heartbeat && heartbeat.isInterrupted())
    {
      logger.info("Startup heartbeat...");
      heartbeat.setInterrupted(false);
      executorService.submit(heartbeat);
    }
    if (null != receiveDetect && receiveDetect.isInterrupted())
    {
      logger.info("Startup receive detect...");
      receiveDetect.setInterrupted(false);
      executorService.submit(receiveDetect);
    }
  }

  private void reconnectSend()
      throws InterruptedException
  {
    if (isReconnect)
    {
      logger.info("Recreate the connection...");
      if (null != send)
        IOUtils.closeQuietly(send);
      Thread.sleep(interval);
      createSendConnection();
    }
  }

  public void close()
  {
    logger.info("Close the connection...");
    isReady = false;
    if (null != heartbeat)
      heartbeat.interrupted();
    if (null != receiveDetect)
      receiveDetect.interrupted();
    IOUtils.closeQuietly(send);
    IOUtils.closeQuietly(receive);
  }

  public void shutdown()
  {
    logger.info("Shutdown the link...");
    close();
    executorService.shutdownNow();
  }

  public void rebuild()
      throws IOException, InterruptedException
  {
    close();
    if (isReconnect)
    {
      logger.info("Reconnection...");
      build();
    }
  }

  public void build()
      throws IOException, InterruptedException
  {
    logger.info("Build the connection...");
    isReady = false;
    if (isReceiveFirst)
    {
      createReceiveConnection();
      createSendConnection();
    }
    else
    {
      createSendConnection();
      try
      {
        createReceiveConnection();
      }
      catch (SocketTimeoutException e)
      {
        logger.error("Create receive connection timeout.", e);
        reconnectSend();
      }
    }

    startDetect();

    if (isConnected(send) && isConnected(receive))
      isReady = true;
  }

  private boolean isConnected(Socket s)
  {
    return null != s && !s.isClosed() && s.isConnected();
  }

  public boolean isReady()
  {
    return isReady;
  }

  private void testReady()
      throws IOException
  {
    if (!isReady)
    {
      logger.error("Connection is not established.");
      throw new IOException("Connection is not established.");
    }
  }

  private void createSendConnection()
      throws InterruptedException
  {
    if (!isReady)
    {
      logger.info("Create send connection...");
      try
      {
        send = new Socket();
        send.connect(new InetSocketAddress(data.getSendIP(), data.getSendPort()), timeout);
        // send.setKeepAlive(true);
      }
      catch (IOException e)
      {
        logger.error("Create send connection error.", e);
        reconnectSend();
      }
    }
  }

  private void createReceiveConnection()
      throws IOException
  {
    if (!isReady)
    {
      logger.info("Waiting receive connection...");
      ServerSocket ss = null;
      try
      {

        ss = new ServerSocket();
        if (!isReceiveFirst)
          ss.setSoTimeout(timeout);
        ss.setReuseAddress(true);
        ss.bind(new InetSocketAddress(data.getRecevicePort()));
        receive = ss.accept();
      }
      catch (IOException e)
      {
        logger.error("Create receive connection error.", e);
        throw e;
      }
      finally
      {
        if (ss != null)
          IOUtils.closeQuietly(ss);
      }
    }
  }

  public void setTimeout(int timeout)
  {
    this.timeout = timeout;
  }

  public void setInterval(long interval)
  {
    this.interval = interval;
  }

  public void write(byte[] data)
      throws IOException
  {
    testReady();
    IOUtils.write(data, send.getOutputStream());
    logger.info("Write data: " + new String(data));
  }

  public byte[] read(int len)
      throws IOException
  {
    testReady();
    byte[] buffer = new byte[len];
    IOUtils.readFully(receive.getInputStream(), buffer);
    logger.info("Read data: " + new String(buffer));
    return buffer;

  }

}
