package com.brilliance.eibs.util;

import com.brilliance.eibs.core.service.Context;
import com.brilliance.eibs.el.CommonFunctionUtils;
import com.brilliance.eibs.util.logging.AbstractLoggingFrame;
import com.brilliance.eibs.util.logging.LoggingFrame;
import com.brilliance.eibs.util.logging.log4j.Log4JLoggingFrame;
import com.brilliance.eibs.util.logging.log4j2.Log4J2LoggingFrame;
import com.brilliance.eibs.util.logging.logback.LogbackLoggingFrame;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.MDC;

import java.io.*;
import java.net.URL;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * 日志工具类
 *
 * @author hujun
 */
public class LogUtil {
    public static final String SERVICE = "services";
    public static final String LOG_SUFFIX = ".log";
    public static final String LOG_PATH = "log";
    public static final String LOG_FILE_NAME_SYMBOL = "interface_id";
    public static boolean mode = false;
    public static final String ROOT_PATH = CommonFunctionUtils.getFilePath("") + "/";
    public static final String LOG_MODE = PropertyUtil.getProperty("sys.properties", "LOGMODE");

    static {
        System.setProperty("root.path", ROOT_PATH);
        if (!StringUtil.isEmpty(LOG_MODE) && "no".equals(LOG_MODE)) {
            mode = true;
        }
        initialize();
    }

    /**
     * 根据传入的字符串创建log对象
     *
     * @param name
     * @return
     */
    public static Logger getLogger(String name) {
        if (mode) {
            return LoggerFactory.getLogger(name);
        } else {
            synchronized (name.intern()) {
                MDC.put(LOG_FILE_NAME_SYMBOL, name);
                return LoggerFactory.getLogger(name);
            }
        }
    }

    public static Logger getOrCreateLogger(String name) {
        Object parentLog = MDC.get(LogUtil.LOG_FILE_NAME_SYMBOL);
        if (parentLog != null) {
            if (parentLog instanceof String) {
                String parentLogName = (String) parentLog;
                if (StringUtil.isNotEmpty(parentLogName)) {
                    return getLogger(parentLogName);
                }
            }
        }
        return getLogger(name);
    }

    /**
     * 重载getLogger方法，可以根据传入的class的获取相应的对象
     *
     * @param clz
     * @return
     */
    public static Logger getLogger(Class<?> clz) {
        return getLogger(clz.getSimpleName());
    }

    public static Logger getLogger(Context context) {
        if (context != null && context.getTransactionMap() != null && context.getTransactionMap().get(Constants.Log_VAR) != null) {
            return (Logger) context.getTransactionMap().get(Constants.Log_VAR);
        }
        return LoggerFactory.getLogger(LogUtil.class);
    }

    public static Logger getTrnLog(String interfaceName, String transactionName, String pre,
                                   boolean merge) {
        if (mode) {
            return LoggerFactory.getLogger(interfaceName + "_" + transactionName);
        }

        Logger logger = null;
        String logFilename = "";
        if (merge && interfaceName.equals(transactionName)) {
            logFilename = interfaceName;
        } else {
            logFilename = interfaceName + "_" + transactionName;
        }
        if (!StringUtil.isEmpty(pre)) {
            logFilename = pre + "_" + logFilename;
        }
        logger = getLogger(logFilename);
        return logger;
    }

    public static Logger getTrnLog(String interfaceName, String transactionName) {
        return getTrnLog(interfaceName, transactionName, null, false);
    }

    public static Logger getTrnLog(String date, String interfaceName, String transactionName) {

        if (mode) {
            return LoggerFactory.getLogger(interfaceName + "_" + transactionName);
        }
        Logger logger = LogUtil.getLogger(interfaceName + "_" + transactionName);
        return logger;
    }


    private static void initialize() {
        try {
            System.setProperty("root.path", AbstractLoggingFrame.toURI(AbstractLoggingFrame.getURL("")).getPath());
            System.setErr(new PrintStream(new BufferedOutputStream(new FileOutputStream(new File(System.getProperty("root.path"), "error.txt"))), true));
            LoggingFrame loggingFrame = findSuitableLoggingFrame();
            // 增加接口平台的根路径作为日志归档的根路径
            loggingFrame.beforeInitialize();
            loggingFrame.initialize();
        } catch (Exception e) {
            System.err.println("日志框架初始化异常:" + e.getMessage());
        } finally {
            // 重新切回
            System.setErr(new PrintStream(new BufferedOutputStream(new FileOutputStream(FileDescriptor.err)), true));
        }
    }


    private static LoggingFrame findSuitableLoggingFrame() {
        Set<URL> staticLoggerBinderPathSet = new LinkedHashSet<URL>();
        try {
            ClassLoader loggerFactoryClassLoader = LogUtil.class.getClassLoader();
            Enumeration<URL> paths;
            if (loggerFactoryClassLoader == null) {
                paths = ClassLoader.getSystemResources("org/slf4j/impl/StaticLoggerBinder.class");
            } else {
                paths = loggerFactoryClassLoader.getResources("org/slf4j/impl/StaticLoggerBinder.class");
            }
            while (paths.hasMoreElements()) {
                URL path = paths.nextElement();
                staticLoggerBinderPathSet.add(path);
            }
        } catch (IOException ioe) {
            throw new RuntimeException("在类路径下获取日志框架资源时出错");
        }
        if (staticLoggerBinderPathSet.size() > 1) {
            System.err.println("发现多个日志框架:" + staticLoggerBinderPathSet);

        } else if (staticLoggerBinderPathSet.size() == 0) {
            throw new RuntimeException("在类路径下未发现日志框架!");
        }
        URL url = Arrays.asList(staticLoggerBinderPathSet.toArray(new URL[0])).get(0);
        LoggingFrame loggingFrame = null;

        if (url.toString().contains("slf4j-log4j12")) {
            // log4j
            loggingFrame = new Log4JLoggingFrame(LogUtil.class.getClassLoader());
        } else if (url.toString().contains("log4j-slf4j-impl")) {
            // log4j2
            loggingFrame = new Log4J2LoggingFrame(LogUtil.class.getClassLoader());
        } else if (url.toString().contains("logback-classic")) {
            // logback
            loggingFrame = new LogbackLoggingFrame(LogUtil.class.getClassLoader());
        } else if (loggingFrame == null) {
            throw new RuntimeException("日志初始化失败,请添加对应的日志框架jar");
        }
        return loggingFrame;
    }


}
