package com.brilliance.eibs.util;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StringField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.*;
import org.apache.lucene.index.IndexWriterConfig.OpenMode;
import org.apache.lucene.queryparser.classic.ParseException;
import org.apache.lucene.queryparser.classic.QueryParser;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TopDocs;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;
import org.apache.lucene.store.RAMDirectory;
import org.apache.lucene.util.Version;

import java.io.*;
import java.nio.charset.Charset;
import java.text.DecimalFormat;
import java.util.*;

public class LuceneUtil {
    public static String INDEX_DIR = "lucene_index_cache";
    private static final Charset UTF_8 = Charset.forName("utf-8");

    /**
     * @param args
     * @throws IOException
     */
    public static void main(String[] args) throws Exception {
        List<Map<String, Object>> srcData = new ArrayList<Map<String, Object>>();
        Map<String, Object> tmp = new HashMap<String, Object>();
        tmp.put("id", "5");
        tmp.put("name", "Aziz");
        tmp.put("address", "wuhan");
        srcData.add(tmp);
        tmp = new HashMap<String, Object>();
        tmp.put("id", "2");
        tmp.put("name", "lifang");
        tmp.put("address", "beijing");
        srcData.add(tmp);
        tmp = new HashMap<String, Object>();
        tmp.put("id", "3");
        tmp.put("name", "liaofang");
        tmp.put("address", "beijing");
        srcData.add(tmp);


        saveIndex(srcData);
        List<Map<String, Object>> rs = findIndexbyName("li");
        System.out.println(rs);

        // 计算
        rs = findBfbById("xiaofang##beijing##30##1233", srcData);
        System.out.println(rs);
    }

    /**
     * 对指定目录进行内存方式的索引，并通过搜索内容返回检索到的文档路径
     *
     * @param docpath   需要搜索的文件目录
     * @param queryword 带搜索的内容
     * @return
     */
    @SuppressWarnings("unchecked")
    public static List<String> QueryFilesNameByRam(String docpath,
                                                   String queryword) {
        Map<String, Object> rs = QueryFilesByRam(docpath, queryword);
        return (List<String>) (rs.containsKey("files") ? rs.get("files")
                : new ArrayList<String>());
    }

    /**
     * 对指定目录进行硬盘方式索引，并通过搜索内容返回检索到的文档路径
     *
     * @param docpath   需要搜索的文件目录
     * @param queryword 带搜索的内容
     * @return
     */
    @SuppressWarnings("unchecked")
    public static List<String> QueryFilesNameByIndexFile(String docpath,
                                                         String queryword) {
        Map<String, Object> rs = QueryFilesByIndexFile(docpath, null, queryword);
        return (List<String>) (rs.containsKey("files") ? rs.get("files")
                : new ArrayList<String>());
    }

    /**
     * 基于硬盘索引方式的根据输入的搜索内容，返回搜索结果集.<br/>
     * 结果集为Map格式，结果集为Map，其中内容为<strong>{
     * 'dsp':描述内容,'hits':命中数目,'files'：查询到的文件路径信息列表}</strong>。
     *
     * @param docpath   需要搜索的文件目录
     * @param queryword 带搜索的内容
     * @return
     */
    public static Map<String, Object> QueryFilesByIndexFile(String docpath,
                                                            String queryword) {
        return QueryFilesByIndexFile(docpath, null, queryword);
    }

    /**
     * 基于内存索引方式的根据输入的搜索内容，返回搜索结果集.<br/>
     * 结果集为Map格式，结果集为Map，其中内容为<strong>{
     * 'dsp':描述内容,'hits':命中数目,'files'：查询到的文件路径信息列表}</strong>。
     *
     * @param docpath   需要搜索的文件目录
     * @param queryword 带搜索的内容
     * @return
     */
    public static Map<String, Object> QueryFilesByRam(String docpath,
                                                      String queryword) {
        try {
            File dir = getDataFile(docpath);
            // 首先建立索引,这里使用内存方式索引
            Directory directory = indexRamMain(dir.getPath());
            // 最后搜索结果
            return SearchFiles(new String[]{"-query", queryword}, directory);
        } catch (Exception e) {
            e.printStackTrace();
            return new HashMap<String, Object>();
        }
    }

    public static Map<String, Object> QueryFilesByIndexFile(String docpath,
                                                            String indexpath, String queryword) {
        String indexpth = INDEX_DIR;
        if (!StringUtil.isEmpty(indexpath))
            indexpth = indexpath;
        try {
            File dir = getDataFile(docpath);
            File indexDir = getIndexDirPath(indexpth);
            // 首先建立索引,这里使用文件索引
            indexFileMain(new String[]{"-create", "-docs", dir.getPath(),
                    "-index", indexDir.getPath()});
            // 最后搜索结果
            return SearchFiles(new String[]{"-query", queryword, "-index",
                    indexDir.getPath()}, null);
        } catch (Exception e) {
            e.printStackTrace();
            return new HashMap<String, Object>();
        }
    }

    private static Map<String, Object> SearchFiles(String[] args,
                                                   Directory directory) throws IOException, ParseException {
        PrintStream outSave = System.out;
        Map<String, Object> rs = new HashMap<String, Object>();
        try {
            ByteArrayOutputStream bytes = new ByteArrayOutputStream();
            PrintStream fakeSystemOut = new PrintStream(bytes, false, Charset
                    .defaultCharset().name());
            System.setOut(fakeSystemOut);
            String index = "index";
            String field = "contents";
            String queryString = null;
            for (int i = 0; i < args.length; i++) {
                if ("-index".equals(args[i])) {
                    index = args[i + 1];
                    i++;
                } else if ("-query".equals(args[i])) {
                    queryString = args[i + 1];
                    i++;
                }
            }
            IndexReader reader = null;
            if (directory != null)
                reader = DirectoryReader.open(directory);
            else
                reader = DirectoryReader
                        .open(FSDirectory.open(new File(index)));
            IndexSearcher searcher = new IndexSearcher(reader);
            Analyzer analyzer = new StandardAnalyzer();
            QueryParser parser = new QueryParser(field, analyzer);
            Query query = parser.parse(queryString + "*");
            System.out.println("Searching for: " + query.toString(field));
            rs = doSearch(searcher, query);
            fakeSystemOut.flush();
            String dsp = bytes.toString(Charset.defaultCharset().name());
            rs.put("dsp", dsp);
            reader.close();
        } finally {
            System.setOut(outSave);
        }
        return rs;
    }

    private static Map<String, Object> doSearch(IndexSearcher searcher,
                                                Query query) throws IOException {
        Map<String, Object> rs = new HashMap<String, Object>();
        List<String> ls = new ArrayList<String>();
        TopDocs results = searcher.search(query, 1);
        try {
            results = searcher.search(query, results.totalHits);
        } catch (Exception e) {
            rs.put("hits", 0);
            rs.put("files", ls);
            return rs;
        }
        ScoreDoc[] hits = results.scoreDocs;
        int numTotalHits = results.totalHits;
        System.out.println(numTotalHits + " total matching documents");
        rs.put("hits", numTotalHits);

        for (int i = 0; i < numTotalHits; i++) {
            Document doc = searcher.doc(hits[i].doc);
            String path = doc.get("path");
            if (path != null) {
                System.out.println((i + 1) + ". " + path);
                ls.add(path);
            } else {
                System.out
                        .println((i + 1) + ". " + "No path for this document");
            }
        }
        rs.put("files", ls);
        return rs;
    }

    private static File getDataFile(String pth) throws IOException {
        File f = new File(pth);
        if (!f.exists())
            throw new IOException("Cannot find resource: " + pth);
        return f;
    }

    static File getIndexDirPath(String pathstr) throws Exception {
        try {
            File f = new File(pathstr);
            if (!f.exists()) {
                throw new IOException("Cannot find resource: " + pathstr);
            }
            return f;
        } catch (Exception e1) {
            // 创建索引目录
            File f = new File(Thread.currentThread().getContextClassLoader()
                    .getResource(pathstr).toURI());
            if (!f.exists()) {
                if (!f.getParentFile().exists()) {
                    if (!f.getParentFile().mkdirs())
                        throw new IOException("create index parent dir fail: "
                                + f.getParentFile());
                    if (!f.mkdir())
                        throw new IOException("create index dir fail: "
                                + pathstr);
                }
            }
            return f;
        }
    }

    static Directory indexRamMain(String docsPath) throws Exception {
        if (docsPath == null) {
            throw new Exception("Cannot find docsPath");
        }
        final File docDir = new File(docsPath);
        if (!docDir.exists() || !docDir.canRead()) {
            System.out
                    .println("Document directory '"
                            + docDir.getAbsolutePath()
                            + "' does not exist or is not readable, please check the path");
            System.exit(1);
        }
        Date start = new Date();
        try {
            Directory dir = new RAMDirectory();
            Analyzer analyzer = new StandardAnalyzer();
            IndexWriterConfig iwc = new IndexWriterConfig(Version.LATEST,
                    analyzer);
            IndexWriter writer = new IndexWriter(dir, iwc);
            indexDocs(writer, docDir);
            writer.close();
            Date end = new Date();
            System.out.println(end.getTime() - start.getTime()
                    + " total milliseconds");
            return dir;
        } catch (IOException e) {
            throw new Exception(
                    "Document directory '"
                            + docDir.getAbsolutePath()
                            + "' does not exist or is not readable, please check the path");
        }
    }

    static void indexFileMain(String[] args) throws Exception {
        String indexPath = "index";
        String docsPath = null;
        boolean create = true;
        for (int i = 0; i < args.length; i++) {
            if ("-index".equals(args[i])) {
                indexPath = args[i + 1];
                i++;
            } else if ("-docs".equals(args[i])) {
                docsPath = args[i + 1];
                i++;
            } else if ("-update".equals(args[i])) {
                create = false;
            }
        }

        if (docsPath == null) {
            throw new Exception("Cannot find docsPath");
        }

        final File docDir = new File(docsPath);
        if (!docDir.exists() || !docDir.canRead()) {
            throw new Exception(
                    "Document directory '"
                            + docDir.getAbsolutePath()
                            + "' does not exist or is not readable, please check the path");
        }

        Date start = new Date();
        try {
            System.out.println("Indexing to directory '" + indexPath + "'...");
            Directory dir = FSDirectory.open(new File(indexPath));
            Analyzer analyzer = new StandardAnalyzer();
            IndexWriterConfig iwc = new IndexWriterConfig(Version.LATEST,
                    analyzer);
            if (create) {
                // 创建新索引，移除之前所有索引
                iwc.setOpenMode(OpenMode.CREATE);
            } else {
                // 在存在的索引文件附加document
                iwc.setOpenMode(OpenMode.CREATE_OR_APPEND);
            }
            IndexWriter writer = new IndexWriter(dir, iwc);
            indexDocs(writer, docDir);
            writer.close();
            Date end = new Date();
            System.out.println(end.getTime() - start.getTime()
                    + " total milliseconds");
        } catch (IOException e) {
            System.out.println(" caught a " + e.getClass()
                    + "\n with message: " + e.getMessage());
        }
    }

    static void indexDocs(IndexWriter writer, File file) throws IOException {
        // do not try to index files that cannot be read
        if (file.canRead()) {
            if (file.isDirectory()) {
                String[] files = file.list();
                // an IO error could occur
                if (files != null) {
                    for (int i = 0; i < files.length; i++) {
                        indexDocs(writer, new File(file, files[i]));
                    }
                }
            } else {
                FileInputStream fis;
                try {
                    fis = new FileInputStream(file);
                } catch (FileNotFoundException fnfe) {
                    return;
                }
                try {
                    Document doc = new Document();
                    Field pathField = new StringField("path", file.getPath(),
                            Field.Store.YES);
                    doc.add(pathField);
                    doc.add(new TextField("contents", new BufferedReader(
                            new InputStreamReader(fis, UTF_8))));
                    if (writer.getConfig().getOpenMode() == OpenMode.CREATE) {
                        System.out.println("adding " + file);
                        writer.addDocument(doc);
                    } else {
                        System.out.println("updating " + file);
                        writer.updateDocument(new Term("path", file.getPath()),
                                doc);
                    }

                } finally {
                    fis.close();
                }
            }
        }
    }

    public static void indexHMD(IndexWriter writer,
                                List<Map<String, Object>> data) throws IOException {
        for (int i = 0; i < data.size(); i++) {
            Document doc = new Document();
            Set<String> keys = data.get(i).keySet();
            if (keys.isEmpty())
                continue;
            for (String key : keys) {
                Field field = null;
                // StringField默认不分词;TextField默认分词
                if ("id".equals(key)) {
                    field = new StringField(key,
                            ((String) data.get(i).get(key)), Field.Store.YES);
                } else {
                    field = new TextField(key, ((String) data.get(i).get(key)),
                            Field.Store.NO);
                }
                doc.add(field);
            }
            if (writer.getConfig().getOpenMode() == OpenMode.CREATE) {
                writer.addDocument(doc);
            } else {
                writer.updateDocument(
                        new Term("id", (String) data.get(i).get("id")), doc);
            }
        }
    }

    public static void createHMDIndex(String[] args,
                                      List<Map<String, Object>> dataSrc) {
        boolean create = true;
        for (int i = 0; i < args.length; i++) {
            if ("-index".equals(args[i])) {
                INDEX_DIR = args[i + 1];
                i++;
            } else if ("-update".equals(args[i])) {
                create = false;
            }
        }

        Date start = new Date();
        try {
            System.out.println("Indexing to directory '" + INDEX_DIR + "'...");
            Directory dir = FSDirectory.open(new File(INDEX_DIR));
            Analyzer analyzer = new StandardAnalyzer();
            IndexWriterConfig iwc = new IndexWriterConfig(Version.LATEST,
                    analyzer);
            if (create) {
                // 创建新索引，移除之前所有索引
                iwc.setOpenMode(OpenMode.CREATE);
            } else {
                // 在存在的索引文件附加document
                iwc.setOpenMode(OpenMode.CREATE_OR_APPEND);
            }
            IndexWriter writer = new IndexWriter(dir, iwc);
            indexHMD(writer, dataSrc);
            writer.close();
            Date end = new Date();
            System.out.println(end.getTime() - start.getTime()
                    + " total milliseconds");
        } catch (IOException e) {
            System.out.println(" caught a " + e.getClass()
                    + "\n with message: " + e.getMessage());
        }

    }

    public static Map<String, Object> getHMDMap(String[] args,
                                                Directory directory) throws IOException, ParseException {
        PrintStream outSave = System.out;
        Map<String, Object> rs = new HashMap<String, Object>();
        try {
            ByteArrayOutputStream bytes = new ByteArrayOutputStream();
            PrintStream fakeSystemOut = new PrintStream(bytes, false, Charset
                    .defaultCharset().name());
            System.setOut(fakeSystemOut);
            String field = "name";
            String queryString = null;
            for (int i = 0; i < args.length; i++) {
                if ("-index".equals(args[i])) {
                    INDEX_DIR = args[i + 1];
                    i++;
                } else if ("-query".equals(args[i])) {
                    queryString = args[i + 1];
                    i++;
                }
            }
            IndexReader reader = null;
            if (directory != null)
                reader = DirectoryReader.open(directory);
            else
                reader = DirectoryReader.open(FSDirectory.open(new File(
                        INDEX_DIR)));
            IndexSearcher searcher = new IndexSearcher(reader);
            Analyzer analyzer = new StandardAnalyzer();
            QueryParser parser = new QueryParser(field, analyzer);
            Query query = parser.parse(queryString + "*");
            System.out.println("Searching for: " + query.toString(field));
            rs = getHMDSearchMap(searcher, query);
            fakeSystemOut.flush();
            String dsp = bytes.toString(Charset.defaultCharset().name());
            rs.put("dsp", dsp);
            reader.close();
        } finally {
            System.setOut(outSave);
        }
        return rs;
    }

    public static List<Map<String, Object>> getHMDList(String[] args,
                                                       Directory directory) throws IOException, ParseException {
        PrintStream outSave = System.out;
        List<Map<String, Object>> rs = new ArrayList<Map<String, Object>>();
        String dsp = "";
        try {
            ByteArrayOutputStream bytes = new ByteArrayOutputStream();
            PrintStream fakeSystemOut = new PrintStream(bytes, false, Charset
                    .defaultCharset().name());
            System.setOut(fakeSystemOut);
            String field = "name";
            String queryString = null;
            for (int i = 0; i < args.length; i++) {
                if ("-index".equals(args[i])) {
                    INDEX_DIR = args[i + 1];
                    i++;
                } else if ("-query".equals(args[i])) {
                    queryString = args[i + 1];
                    i++;
                }
            }
            IndexReader reader = null;
            if (directory != null)
                reader = DirectoryReader.open(directory);
            else
                reader = DirectoryReader.open(FSDirectory.open(new File(
                        INDEX_DIR)));
            IndexSearcher searcher = new IndexSearcher(reader);
            Analyzer analyzer = new StandardAnalyzer();
            QueryParser parser = new QueryParser(field, analyzer);
            Query query = parser.parse(queryString + "*");
            System.out.println("Searching for: " + query.toString(field));
            rs = getHMDSearchList(searcher, query);
            fakeSystemOut.flush();
            dsp = bytes.toString(Charset.defaultCharset().name());

            reader.close();
        } finally {
            System.setOut(outSave);
            System.out.println(dsp);
        }
        return rs;
    }

    private static Map<String, Object> getHMDSearchMap(IndexSearcher searcher,
                                                       Query query) throws IOException {
        Map<String, Object> rs = new HashMap<String, Object>();
        List<String> ls = new ArrayList<String>();
        TopDocs results = searcher.search(query, 100);
        try {
            results = searcher.search(query, results.totalHits);
        } catch (Exception e) {
            rs.put("hits", 0);
            rs.put("ids", ls);
            return rs;
        }
        ScoreDoc[] hits = results.scoreDocs;
        int numTotalHits = results.totalHits;
        if (numTotalHits > 50) {
            numTotalHits = 50;
        }
        System.out.println(numTotalHits + " total matching documents");
        rs.put("hits", numTotalHits);

        for (int i = 0; i < numTotalHits; i++) {
            Document doc = searcher.doc(hits[i].doc);
            String path = doc.get("id");
            if (path != null) {
                System.out.println((i + 1) + ". " + path);
                ls.add(path);
            } else {
                System.out.println((i + 1) + ". "
                        + "No match for this document");
            }
        }
        rs.put("ids", ls);
        return rs;
    }

    private static List<Map<String, Object>> getHMDSearchList(
            IndexSearcher searcher, Query query) throws IOException {
        List<Map<String, Object>> rs = new ArrayList<Map<String, Object>>();
        TopDocs results = searcher.search(query, 100);
        try {
            results = searcher.search(query, results.totalHits);
        } catch (Exception e) {
            return rs;
        }
        ScoreDoc[] hits = results.scoreDocs;
        int numTotalHits = results.totalHits;
        if (numTotalHits > 50) {
            numTotalHits = 50;
        }
        System.out.println(numTotalHits + " total matching documents");
        for (int i = 0; i < numTotalHits; i++) {
            Document doc = searcher.doc(hits[i].doc);
            String path = doc.get("id");
            if (path != null) {
                Map<String, Object> tmp = new HashMap<String, Object>();
                System.out.println((i + 1) + ". " + path);
                tmp.put("id", path);
                rs.add(tmp);
            } else {
                System.out.println((i + 1) + ". "
                        + "No match for this document");
            }
        }
        return rs;
    }

    /**
     * 基于更新方式的保存文件索引
     *
     * @param dataSrc   数据源
     * @param indexpath
     */
    public static void saveIndex(List<Map<String, Object>> dataSrc) {
        createHMDIndex(new String[]{"-update"}, dataSrc);
    }

    /**
     * 搜索name的索引结果
     *
     * @param name 查询参数
     * @return
     */
    public static List<Map<String, Object>> findIndexbyName(String name) {
        try {
            return getHMDList(new String[]{"-query", name}, null);
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
    }

    /**
     * 查询相似度大于指定百分比的结果集
     *
     * @param content name##address##百分比##机构号
     * @param list    待检查的数据源
     * @return
     */
    public static List<Map<String, Object>> findBfbById(String content,
                                                        List<Map<String, Object>> list) {
        String[] args = content.split("##");
        // name##address##百分比##机构号
        String name = args[0];
        String address = args[1];
        String bfbArgs = args[2];
        // String bchkey = args[3];
        double jqd = Double.parseDouble(bfbArgs) / 100;
        DecimalFormat df = null;
        df = new DecimalFormat("0.0");
        List<Map<String, Object>> rs = new ArrayList<Map<String, Object>>();
        for (int i = 0; i < list.size(); i++) {
            Map<String, Object> element = list.get(i);
            double bfbNum;
            if (element.containsKey("address")
                    && element.get("address") != null) {
                double bfbNum1 = XSDUtil
                        .sim(name, (String) element.get("name"));
                double bfbNum2 = XSDUtil.sim(address,
                        (String) element.get("address"));
                bfbNum = (bfbNum1 + bfbNum2) / 2;
            } else {
                bfbNum = XSDUtil.sim(name, (String) element.get("name"));
            }
            if (bfbNum <= jqd) {
                continue;
            }
            element.put("bfb", df.format(bfbNum * 100) + "%");
            rs.add(element);
        }
        return rs;
    }
}
